from inspect import signature
import types

import pytest

from ioisis.fieldutils import SubfieldParser


SFP_SIGNATURE = signature(SubfieldParser)
SFP_DEFAULT_EMPTY = SFP_SIGNATURE.parameters["empty"].default
SFP_DEFAULT_LENGTH = SFP_SIGNATURE.parameters["length"].default
SFP_DEFAULT_LOWER = SFP_SIGNATURE.parameters["lower"].default

SFP_DATA = {  # Items are {id: (field, expected, kwargs)}
    # Empty input
    "empty_false":
        ("", [], dict(prefix="x")),
    "empty_true":
        ("", [("", "")], dict(prefix="x", empty=True)),

    # Single non-empty subfield (and perhaps some empty subfield)
    "single_nonempty_subfield_first":
        ("data", [("", "data")], dict(prefix="^")),
    "single_nonempty_subfield_no_first":
        ("data", [("a", "ta")], dict(prefix="d")),
    "single_nonempty_subfield_empty_first":
        ("data", [("", ""), ("a", "ta")], dict(prefix="d", empty=True)),

    # Non-subfield prefix (trailing prefix and field named with prefix)
    "non_subfield_prefix":
        ("data", [("", "d"), ("t", "a")], dict(prefix="a")),

    # UTF-8 / multi-byte prefix
    "utf8_prefix":
        ("dátá", [("", "d"), ("t", "á")], dict(prefix="á")),
    "multibyte_ascii_prefix":
        ("#-#ak0#-ak-#", [("#", "ak0"), ("a", "k-#")], dict(prefix="#-")),

    # Length, number and zero
    "length_2_ignore_empty":
        ("data", [("", "d")], dict(prefix="a", length=2)),
    "length_2_keep_empty": (
        "data",
        [("", "d"), ("ta", "")],
        dict(prefix="a", length=2, empty=True),
    ),
    "length_0_ignore_empty":
        ("data", [("", "d"), ("1", "t")], dict(prefix="a", length=0)),
    "length_0_ignore_empty_no_number": (
        "data",
        [("", "d"), ("", "t")],
        dict(prefix="a", length=0, number=False),
    ),
    "length_0_keep_empty": (
        "data",
        [("", "d"), ("1", "t"), ("2", "")],
        dict(prefix="a", length=0, empty=True),
    ),
    "length_0_keep_empty_no_number": (
        "ðata",
        [("", "ð"), ("", "t"), ("", "")],
        dict(prefix="a", length=0, empty=True, number=False),
    ),
    "length_0_keep_empty_zero": (
        "data",
        [("0", "d"), ("1", "t"), ("2", "")],
        dict(prefix="a", length=0, empty=True, zero=True),
    ),

    # First, number and zero
    "first_unused":
        ("ioisis test", [("s", " test")], dict(prefix="i", first="1")),
    "first_empty": (
        "ioisis test",
        [("1", ""), ("o", ""), ("s", ""), ("s1", " test")],
        dict(prefix="i", first="1", empty=True),
    ),
    "first_empty_no_number": (
        "ioisis test",
        [("1", ""), ("o", ""), ("s", ""), ("s", " test")],
        dict(prefix="i", first="1", empty=True, number=False),
    ),
    "first_empty_zero": (
        "ioisis test",
        [("_0", ""), ("o0", ""), ("s0", ""), ("s1", " test")],
        dict(prefix="i", first="_", empty=True, zero=True),
    ),
    "first_with_3_bytes": (
        "ioisis test",
        [("1st", "io"), ("i", "s test")],
        dict(prefix="is", first="1st"),
    ),
    "first_with_3_bytes_and_remaining_with_length_2": (
        "ioisis test",
        [("1st", "io"), ("is", " test")],
        dict(prefix="is", first="1st", length=2),
    ),
    "first_with_3_bytes_and_remaining_with_length_2_number": (
        "ioisis test isis numbered",
        [("1st", "io"), ("is", " test "), ("is1", " numbered")],
        dict(prefix="is", first="1st", length=2),
    ),
    "first_with_3_bytes_and_remaining_with_length_2_number_zero": (
        "ioisis të§t isis numbered",
        [("1st0", "io"), ("is0", " të§t "), ("is1", " numbered")],
        dict(prefix="is", first="1st", length=2, zero=True),
    ),
    "first_with_3_bytes_and_remaining_with_length_2_no_number": (
        "ioisis test isisnt numbered",
        [("1st", "io"), ("is", " test "), ("is", "nt numbered")],
        dict(prefix="is", first="1st", length=2, number=False),
    ),

    # Lower
    "lower_no_number_length_2": (
        "7Asuiñ¼suidn7AIDjqoiw7siojAipoo7Aidosijd",
        [("su", "iñ¼suidn"), ("id", "jqoiw7siojAipoo"), ("id", "osijd")],
        dict(prefix="7A", length=2, lower=True, number=False),
    ),
    "number_no_lower_length_2": (
        "7Asuiñ¼suidn7AIDjqoiw7siojAipoo7Aidosijd",
        [("su", "iñ¼suidn"), ("ID", "jqoiw7siojAipoo"), ("id", "osijd")],
        dict(prefix="7A", length=2, lower=False, number=True),
    ),
    "lower_number_zero_length_2": (
        "7Asuiñ¼suidn7AIDjqoiw7siojAipoo7Aidosijd",
        [("su0", "iñ¼suidn"), ("id0", "jqoiw7siojAipoo"), ("id1", "osijd")],
        dict(prefix="7A", length=2, lower=True, number=True, zero=True),
    ),
    "lower_first_empty": (
        "",
        [("first", "")],
        dict(prefix="^", lower=True, first="FIRST", empty=True),
    ),
    "lower_empty_no_first_number": (
        "^ASOME^1DATA^AHERE",
        [("", ""), ("a", "SOME"), ("1", "DATA"), ("a1", "HERE")],
        dict(prefix="^", lower=True, empty=True, number=True),
    ),
    "lower_no_empty_not_resynthesizable": (
        "^ASOME^1^A^2^a^3DATA^a^AHERE^a¿",
        [("a", "SOME"), ("3", "DATA"), ("a1", "HERE"), ("a2", "¿")],
        dict(prefix="^", lower=True, empty=False, number=True),
    ),
}

# In SFP_DATA, either empty is False, lower is True or field == resynth
SFP_DATA_FIELD_RESYNTH_ASSUMING_EMPTY = {
    id_: expected[0][1] + "".join(
        kwargs["prefix"] + k[:kwargs.get("length", SFP_DEFAULT_LENGTH)] + v
        for k, v in expected[1:]
    )
    for id_, (field, expected, kwargs) in SFP_DATA.items()
    if expected
}
assert all(not kwargs.get("empty", SFP_DEFAULT_EMPTY)
           or kwargs.get("lower", SFP_DEFAULT_LOWER)
           or field == SFP_DATA_FIELD_RESYNTH_ASSUMING_EMPTY[id_]
           for id_, (field, expected, kwargs) in SFP_DATA.items())

# Copy of tests where empty=False but it can also be True
SFP_EXTRA_DATA = {
    id_ + "_with_empty_true": (field, expected, {**kwargs, "empty": True})
    for id_, (field, expected, kwargs) in SFP_DATA.items()
    if expected  # When empty=True, the result has at least one subfield
    and not kwargs.get("empty", SFP_DEFAULT_EMPTY)
    and field == SFP_DATA_FIELD_RESYNTH_ASSUMING_EMPTY[id_]
}
assert SFP_EXTRA_DATA  # We know there is at least one such case

# Build the test params to create the tests for both str and bytes
SFP_TEST_PARAMS_STR = [pytest.param(*v, id=k + "_decoded_str")
                       for k, v in {**SFP_DATA, **SFP_EXTRA_DATA}.items()]
SFP_TEST_PARAMS_BYTES = [
    pytest.param(
        field.encode("utf-8"),
        [(k.encode("utf-8"), v.encode("utf-8")) for k, v in expected],
        {k: v.encode("utf-8") if isinstance(v, str) else v
         for k, v in kwargs.items()},
        id=id_ + "_utf8_encoded_bytes",
    )
    for id_, (field, expected, kwargs)
    in {**SFP_DATA, **SFP_EXTRA_DATA}.items()
]
SFP_TEST_PARAMS = SFP_TEST_PARAMS_STR + SFP_TEST_PARAMS_BYTES


@pytest.mark.parametrize("field, expected, kwargs", SFP_TEST_PARAMS)
def test_sfp_call(field, expected, kwargs):
    sfp = SubfieldParser(**kwargs)
    result = sfp(field)
    assert isinstance(result, types.GeneratorType)
    result_list = list(result)
    assert result_list == expected


@pytest.mark.parametrize("field, expected, kwargs", SFP_TEST_PARAMS)
def test_sfp_unparse_check(field, expected, kwargs):
    sfp = SubfieldParser(**kwargs)
    valid_resynth = sfp.unparse(*expected, check=True)  # Should not raise
    assert isinstance(valid_resynth, type(field))

    # It's resynthesizable when empty is True and lower is False ...
    kwargs_etlf = {**kwargs, "empty": True, "lower": False}
    sfp_etlf = SubfieldParser(**kwargs_etlf)
    unexpected = list(sfp_etlf(field))  # Full data for resynthesis
    assert sfp_etlf.unparse(*unexpected) == field

    # ... as well as when these options makes no difference
    rf_empty = sfp.empty or all(v for k, v in unexpected[1:])
    rf_lower = not sfp.lower or all(k == k.lower() for k, v in unexpected[1:])
    is_resynthesizable = rf_empty and rf_lower
    assert is_resynthesizable == (field == valid_resynth)

    # We can also use the sfp.unparse(*unexpected),
    # mixing the SubfieldParser instance and subfields pairs
    # to resynthesize the field data, but that might raise on checking:
    # It works when the field is resynthesizable
    # and its first key is matching
    has_first = expected and sfp.fz == expected[0][0]
    is_first_case_lower_valid = kwargs.get("first", sfp.first) == sfp.first
    can_mix = is_resynthesizable and is_first_case_lower_valid
    can_check = can_mix and has_first
    assert sfp.unparse(*unexpected, check=can_check) == valid_resynth
    if not has_first:
        assert sfp.unparse(*unexpected[1:], check=can_mix) == valid_resynth
    if not can_check:
        with pytest.raises(ValueError):
            sfp.unparse(*unexpected, check=True)



from collections import defaultdict
from contextlib import closing
from hashlib import sha256
import os
from urllib.request import urlopen

from .java import generator_blocking_process, jvm
from .mst import DEFAULT_MST_ENCODING


BRUMA_URL = "https://github.com/scieloorg/isis2json/raw/def7327/lib/Bruma.jar"
BRUMA_JAR = os.path.join(os.path.expanduser("~"), ".ioisis", "Bruma.jar")
BRUMA_HASH = "a68c3f21ad98a21de49b2eb1c75ccc63bb46f336c1b285ea7fc6d128d29fc57d"


class BrumaCheckError(Exception):
    pass


def check_bruma():
    """Check if Bruma.jar is valid, and download it if it's missing."""
    if not os.path.exists(BRUMA_JAR):
        download_bruma()
    if not os.path.isfile(BRUMA_JAR):
        raise BrumaCheckError("Bruma.jar isn't a file")
    with open(BRUMA_JAR, "rb") as raw_bruma_file:
        if sha256(raw_bruma_file.read()).hexdigest() != BRUMA_HASH:
            raise BrumaCheckError("Invalid Bruma.jar file")


def download_bruma():
    """Simply download BRUMA_URL to BRUMA_JAR."""
    os.makedirs(os.path.dirname(BRUMA_JAR), exist_ok=True)
    with open(BRUMA_JAR, "wb") as raw_bruma_file:
        raw_bruma_file.write(urlopen(BRUMA_URL).read())


@generator_blocking_process
def iter_records(mst_filename, encoding=DEFAULT_MST_ENCODING):
    check_bruma()
    with jvm(domains=["bruma"], classpath=BRUMA_JAR):
        from bruma.master import MasterFactory, Record
        mf = MasterFactory.getInstance(mst_filename).setEncoding(encoding)
        with closing(mf.open()) as mst:
            for record in mst:
                result = defaultdict(list)
                result["active"] = record.getStatus() == Record.Status.ACTIVE
                result["mfn"] = record.getMfn()
                for field in record.getFields():
                    tag = field.getId()
                    result[tag].append(field.getContent())
                yield result


@generator_blocking_process
def iter_tl(
    mst_filename,
    encoding=DEFAULT_MST_ENCODING,
    only_active=True,
    prepend_mfn=False,
    prepend_status=False,
):
    check_bruma()
    with jvm(domains=["bruma"], classpath=BRUMA_JAR):
        from bruma.master import MasterFactory
        mf = MasterFactory.getInstance(mst_filename).setEncoding(encoding)
        with closing(mf.open()) as mst:
            for record in mst:
                status_name = record.getStatus().name()
                if status_name != "ACTIVE" and (only_active or
                                                status_name == "PHYDEL"):
                    continue
                result = []
                if prepend_mfn:
                    result.append(("mfn", "%d" % record.getMfn()))
                if prepend_status:
                    status = {"ACTIVE": "0", "LOGDEL": "1"}[status_name]
                    result.append(("status", status))
                for field in record.getFields():
                    result.append((field.getId(), field.getContent()))
                yield result



"""Custom construct subclasses."""
from contextlib import closing

from construct import Adapter, Array, Check, RepeatUntil, Struct, Subconstruct

from .streamutils import LineSplittedBytesStreamWrapper


DEFAULT_LINE_LEN = 80
DEFAULT_NEWLINE = b"\n"


class DictSegSeq(Adapter):

    def __init__(self, idx_field, subcon, block_size, empty_item,
                 check_nonempty):
        super().__init__(IndexedRange(idx_field, Array(block_size, subcon)))
        self.block_size = block_size
        self.empty_item = empty_item
        self.check_nonempty = check_nonempty

    def _decode(self, obj, context, path):
        return {
            bidx * self.block_size + idx: item
            for bidx, blk in enumerate(obj)
            for idx, item in enumerate(blk, 1)
            if self.check_nonempty(item)
        }

    def _encode(self, obj, context, path):
        return [
            [obj.get(idx, self.empty_item)
             for idx in range(start, start + self.block_size)]
            for start in range(1, max(obj) + 1, self.block_size)
        ]


class IndexedRange(Adapter):
    """Like GreedyRange, but prefixed with an index starting from 1,
    and whose sign is a flag that tells, when negative,
    that we're dealing with the last item.
    """
    def __init__(self, idx_field, subcon):
        super().__init__(RepeatUntil(
            lambda obj, values, ctx: obj["idx"] < 0,
            Struct(
                "idx" / idx_field,
                Check(lambda this: abs(this.idx) == this._index + 1),
                "data" / subcon,
            ),
        ))

    def _decode(self, obj, context, path):
        return [el.data for el in obj]

    def _encode(self, obj, context, path):
        last_idx = len(obj)
        return [{
            "idx": -idx if idx == last_idx else idx,
            "data": el,
        } for idx, el in enumerate(obj, 1)]


class IntInASCII(Adapter):
    """Adapter for Bytes to use it as ASCII numbers."""
    def _decode(self, obj, context, path):
        return int(obj, base=10)

    def _encode(self, obj, context, path):
        length = self.subcon.sizeof(**context)
        return (b"%d" % obj).zfill(length)


class LineSplitRestreamed(Subconstruct):
    """Alternative to Restreamed
    that parses a "line splitted" data,
    builds the lines appending the ``newline`` character/string,
    and works properly with a last incomplete chunk.
    """
    def __init__(self, subcon, line_len=DEFAULT_LINE_LEN,
                 newline=DEFAULT_NEWLINE):
        super().__init__(subcon)
        self.line_len = line_len
        self.newline = newline

    def _parse(self, stream, context, path):
        with closing(LineSplittedBytesStreamWrapper(
            substream=stream,
            line_len=self.line_len,
            newline=self.newline,
        )) as stream2:
            return self.subcon._parsereport(stream2, context, path)

    def _build(self, obj, stream, context, path):
        with closing(LineSplittedBytesStreamWrapper(
            substream=stream,
            line_len=self.line_len,
            newline=self.newline,
        )) as stream2:
            self.subcon._build(obj, stream2, context, path)
        return obj

    def _sizeof(self, context, path):
        n = self.subcon._sizeof(context, path)
        return n + (n // self.line_len + 1) * len(self.newline),


class Unnest(Adapter):
    """Adapter for dict-like containers to unnest (embed) substructures."""
    def __init__(self, names, subcon):
        super().__init__(subcon)
        self.names = list(names)

    def _decode(self, obj, context, path):
        result = obj.copy()
        for name in self.names:
            if name in result:
                result.update(result.pop(name))
        return result

    def _encode(self, obj, context, path):
        result = obj.copy()
        for name in self.names:
            result[name] = obj
        return result



import io

from ioisis.iso import con2dict, DEFAULT_RECORD_STRUCT, \
                       iter_raw_tl, iter_records
from ioisis.fieldutils import nest_decode, tl2record


def test_tag_zero():
    iso_data = DEFAULT_RECORD_STRUCT.build({
        "dir": [{"tag": b"000"}],
        "fields": [b"data"],
    })
    con = DEFAULT_RECORD_STRUCT.parse(iso_data)
    expected = {"0": ["data"]}
    assert con2dict(con, encoding="ascii") == expected


def test_converting_iter_raw_tl_result_to_behave_like_iter_records():
    iso_data = DEFAULT_RECORD_STRUCT.build({
        "dir": [{"tag": b"100"}, {"tag": b"001"}, {"tag": b"010"}],
        "fields": ["™©®".encode("utf-8"), b"data ", b" here"],
    })
    tl, = iter_raw_tl(io.BytesIO(iso_data))
    record, = iter_records(io.BytesIO(iso_data), encoding="utf-8")
    assert record == nest_decode(tl2record(tl), encoding="utf-8")



from codecs import escape_decode
from functools import reduce
from inspect import signature
from io import BytesIO
import signal
import sys

import click
import ujson

from . import bruma, iso, mst
from .fieldutils import nest_decode, nest_encode, SubfieldParser, \
                        tl2record, record2tl, utf8_fix_nest_decode


DEFAULT_JSONL_ENCODING = "utf-8"
INPUT_PATH = object()
OUTPUT_PATH = object()


def apply_decorators(*decorators):
    """Decorator that applies the decorators in reversed order."""
    return lambda func: reduce(lambda f, d: d(f), decorators[::-1], func)


def option(*args, **kwargs):
    """Same to click.option, but saves the args/kwargs for filtering."""
    result = click.option(*args, **kwargs)
    result.args = args
    result.kwargs = kwargs
    return result


def encoding_option(file_ext, default, **kwargs):
    ctx_attr = file_ext + "_encoding"
    return click.option(
        ctx_attr,
        f"--{file_ext[0]}enc",
        default=default,
        show_default=True,
        callback=lambda ctx, param, value:
            setattr(ctx, ctx_attr, value) or value,
        is_eager=True,
        help=f"{file_ext.upper()} file encoding.",
        **kwargs,
    )


def file_arg_enc_option(file_ext, mode, default_encoding):
    arg_kwargs = {}
    if mode is INPUT_PATH:
        arg_name = file_ext + "_input"
        arg_kwargs["type"] = click.Path(
            dir_okay=False,
            resolve_path=True,
            allow_dash=False,
            readable=True,
            writable=False,
        )
    elif mode is OUTPUT_PATH:
        arg_name = file_ext + "_output"
        arg_kwargs["type"] = click.Path(
            dir_okay=False,
            resolve_path=True,
            allow_dash=False,
            readable=False,
            writable=True,
        )
    else:
        arg_kwargs["default"] = "-"
        if "w" in mode:
            arg_name = file_ext + "_output"
        else:
            arg_name = file_ext + "_input"
        if "b" in mode:
            arg_kwargs["type"] = click.File(mode)
        else:
            ctx_attr = file_ext + "_encoding"
            arg_kwargs["callback"] = lambda ctx, param, value: \
                click.File(mode, encoding=getattr(ctx, ctx_attr))(value)

    return apply_decorators(
        encoding_option(file_ext, default=default_encoding),
        click.argument(arg_name, **arg_kwargs)
    )


def iso_bytes_option_with_default(*args, **kwargs):
    return option(
        *args,
        metavar="BYTES",
        show_default=True,
        callback=lambda ctx, param, value:
            escape_decode(value.encode(ctx.iso_encoding))[0],
        **kwargs
    )


def kw_call(func, *args, **kwargs):
    """Call ``func`` without the extra unknown keywords."""
    sig_keys = signature(func).parameters.keys()
    return func(*args, **{k: kwargs[k] for k in sig_keys if k in kwargs})


iso_options = [
    iso_bytes_option_with_default(
        "field_terminator", "--ft",
        default=iso.DEFAULT_FIELD_TERMINATOR,
        help="ISO Field terminator",
    ),
    iso_bytes_option_with_default(
        "record_terminator", "--rt",
        default=iso.DEFAULT_RECORD_TERMINATOR,
        help="ISO Record terminator",
    ),
    option(
        "line_len", "--line",
        default=iso.DEFAULT_LINE_LEN,
        show_default=True,
        help="Line size to wrap the raw ISO data into several lines. "
             "If zero, performs no line splitting.",
    ),
    iso_bytes_option_with_default(
        "newline", "--eol",
        default=iso.DEFAULT_NEWLINE,
        help="End of line character/string for ISO line splitting. "
             "Ignored if --line=0.",
    ),
]


jsonl_mode_option = click.option(
    "--mode", "-m",
    type=click.Choice(["field", "pairs", "nest"], case_sensitive=False),
    default="field",
    help="Mode of JSONL record structure processing "
         "and of field/subfield parsing.",
)


subfield_options = [
    option(
        "--prefix",
        metavar="BYTES",
        default=b"^",
        show_default=True,
        callback=lambda ctx, param, value:
            escape_decode(value.encode("ascii"))[0],
        help="Subfield prefix mark."
    ),
    option(
        "--length",
        default=1,
        show_default=True,
        help="Subfield key length in bytes."
    ),
    option(
        "--lower/--no-lower",
        default=True,
        show_default=True,
        help="Put subfield keys in lower case, "
             "making them case insensitive."
    ),
    option(
        "--first",
        metavar="BYTES",
        default=b"_",
        show_default=True,
        callback=lambda ctx, param, value:
            escape_decode(value.encode("ascii"))[0],
        help="Key to be used for the first keyless subfield."
    ),
    option(
        "--empty/--no-empty",
        default=False,
        show_default=True,
        help="Keep subfield pairs with empty values."
    ),
    option(
        "--number/--no-number",
        default=True,
        show_default=True,
        help="Add a number suffix to the subfield keys "
             "that have already appeared before for the same field. "
             "The suffix numbers start in 1."
    ),
    option(
        "--zero/--no-zero",
        default=False,
        show_default=True,
        help='Add the "0" suffix '
             "to the first of each distinct subfield key. "
             "Has no effect when --no-number."
    ),
]


subfield_unparse_check_option = click.option(
    "--sfcheck/--no-sfcheck",
    default=True,
    show_default=True,
    help="Check if the subfield unparsing rules match its contents, "
         "that is, check if the field generated from the subfields "
         "would produce the same subfields."
)


metadata_filtering_options = [
    option(
        "only_active", "--only-active/--all",
        default=True,
        show_default=True,
        help='Select only records whose status is "0" (active).',
    ),
    option(
        "--prepend-mfn/--no-mfn",
        default=False,
        show_default=True,
        help='Prepend the "mfn" field.',
    ),
    option(
        "--prepend-status/--no-status",
        default=False,
        show_default=True,
        help='Prepend the "status" field from the record metadata, '
             "whose value is either "
             '"0" (active) or "1" (logically deleted).',
    ),
]


mst_options = [
    option(
        "endianness", "--end",
        default=mst.DEFAULT_ENDIANNESS,
        show_default=True,
        type=click.Choice(["little", "big"]),
        help="Byte order endianness for 16/32 bits integer numbers. "
             'Little endian is known as "swapped" in CISIS/Bruma.',
    ),
    option(
        "endianness", "--le",
        flag_value="little",
        is_eager=True,
        help="Same to --end=little.",
    ),
    option(
        "endianness", "--be",
        flag_value="big",
        is_eager=True,
        help="Same to --end=big.",
    ),
    option(
        "--format",
        default=mst.DEFAULT_FORMAT,
        show_default=True,
        type=click.Choice(["isis", "ffi"]),
        help="Leader and directory format mode in master files. "
             "In the ISIS format mode, "
             "the addressing and length fields "
             "(MFRL, BASE, POS and LEN) have 2 bytes, "
             "whereas in the FFI format mode they have 4 bytes.",
    ),
    option(
        "format", "--isis",
        flag_value="isis",
        is_eager=True,
        help="Same to --format=isis.",
    ),
    option(
        "format", "--ffi",
        flag_value="ffi",
        is_eager=True,
        help="Same to --format=ffi.",
    ),
    option(
        "--lockable/--no-locks",
        default=mst.DEFAULT_LOCKABLE,
        show_default=True,
        help="Multi-user locking, "
             "where the MFRL sign is a RLOCK (record lock flag), "
             "MFCXX2 is the DELOCK (Data entry lock / RLOCK counter) "
             "and MFCXX3 is the EWLOCK (Exclusive write lock). "
             "MFRL will be interpreted as unsigned if --no-locks, "
             "effectively increasing the maximum record size "
             "to twice plus one.",
    ),
    option(
        "default_shift", "--shift",
        default=mst.DEFAULT_SHIFT,
        show_default=True,
        help="MSTXL value, the number of XRF bit shift steps. "
             "It affects the minimum possible modulus "
             "for record alignment in the MST file, "
             "and it's part of the control record. "
             "To get the standard ISIS behavior, this should be 0, "
             "and to get the CISIS FFI behavior, this should be 3, "
             "although 6 is the most common choice for large files "
             "in both format modes.",
    ),
    option(
        "--shift4is3/--shift4isnt3",
        default=mst.DEFAULT_SHIFT4IS3,
        show_default=True,
        help="Legacy shifting interpretation "
             "where there's no 4 bits shifting, "
             "and MSTXL=4 should be replaced by MSTXL=3.",
    ),
    option(
        "--min-modulus",
        default=mst.DEFAULT_MIN_MODULUS,
        show_default=True,
        help="Smallest modulus value for record alignment. "
             "Due to XRF shifting, the actual modulus is 2**MSTXL "
             "(2 raised to the power of MSTXL), "
             "unless this value is higher than that. "
             "The ISIS standard define this as 2. "
             "This option makes it possible "
             "to disable the 2 bytes (WORD) alignment of records "
             "by setting it as 1.",
    ),
    option(
        "--packed/--unpacked",
        default=mst.DEFAULT_PACKED,
        show_default=True,
        help="Control the leader and FFI directory alignment. "
             "If --packed, "
             "there should be no padding/filler/slack bytes in these. "
             "If --unpacked, there should be a 4-bytes alignment, "
             "adding 2 filler bytes in the leader "
             "(after the MFRL in ISIS, after the BASE in FFI) "
             "and 2 filler bytes after the TAG "
             "in FFI directory entries. "
             "These are also known as align0/align2 in Bruma "
             "(i.e., with the number of padding bytes), "
             "and as PC/LINUX (or Windows/Linux) in CISIS "
             "(because it used to be compiled in Windows "
             " with -fpack-struct=1, and without it in Linux), "
             "though this option has nothing to do operating system.",
    ),
    option(
        "--filler",
        metavar="HEX_BYTE",
        default="%02X" % ord(mst.DEFAULT_FILLER),
        show_default=True,
        callback=lambda ctx, param, value:
            bytes([int(value, 16)]) if value else None,
        help="Character code in hexadecimal for unset filler options "
             "that doesn't have a specific default.",
    ),
    option(
        "--control-filler",
        metavar="HEX_BYTE",
        callback=lambda ctx, param, value:
            bytes([int(value, 16)]) if value else None,
        help="Filler character code "
             "for the trailing bytes of the control record. "
             "The CISIS source code tells "
             'it should be "FF" for Unisys, and "00" otherwise.',
    ),
    option(
        "--slack-filler",
        metavar="HEX_BYTE",
        callback=lambda ctx, param, value:
            bytes([int(value, 16)]) if value else None,
        help="Filler character code for alignment "
             "in the leader and the directory of all records. "
             "Has no effect when --packed.",
    ),
    option(
        "--block-filler",
        metavar="HEX_BYTE",
        callback=lambda ctx, param, value:
            bytes([int(value, 16)]) if value else None,
        help="Filler character code "
             "for the trailing recordless bytes of a block.",
    ),
    option(
        "--record-filler",
        metavar="HEX_BYTE",
        default="%02X" % ord(mst.DEFAULT_RECORD_FILLER),
        show_default=True,
        callback=lambda ctx, param, value:
            bytes([int(value, 16)]) if value else None,
        help="Filler character code for the trailing record data.",
    ),
    option(
        "--control-len",
        default=mst.DEFAULT_CONTROL_LEN,
        show_default=True,
        help="Control record length, at least 32. "
             "It must be multiple of the modulus.",
    ),
]

mst_ibp_option = click.option(
    "--ibp",
    type=click.Choice(["check", "ignore", "store"]),
    default=mst.DEFAULT_IBP,
    help="Invalid block padding content/size, "
         "which might appear as a residual of some previous content "
         "when the MST file gets updated in place. "
         'If "store", the contents are stored in the previous record.',
)


utf8_fix_option = click.option(
    "utf8_fix", "--utf8",
    is_flag=True,
    default=False,
    help="Decode the input data with UTF-8 if possible, "
         "using the given input encoding as a fallback.",
)


@click.group()
def main():
    """ISIS data converter using the ioisis Python library."""
    try:  # Fix BrokenPipeError by opening a new fake standard output
        signal.signal(signal.SIGPIPE,
                      lambda signum, frame: setattr(sys, "stdout", BytesIO()))
    except (AttributeError, ValueError):
        pass  # No SIGPIPE in this OS


@main.command("bruma-mst2jsonl")
@apply_decorators(*metadata_filtering_options)
@jsonl_mode_option
@apply_decorators(*subfield_options)
@file_arg_enc_option("mst", INPUT_PATH, mst.DEFAULT_MST_ENCODING)
@file_arg_enc_option("jsonl", "w", DEFAULT_JSONL_ENCODING)
def bruma_mst2jsonl(mst_input, jsonl_output, mst_encoding, mode, **kwargs):
    """MST+XRF to JSON Lines based on a Bruma (requires Java)."""
    ensure_ascii = jsonl_output.encoding.lower() == "ascii"
    kwargs_menc = {key: kwargs[key].decode(mst_encoding)
                   for key in ["prefix", "first"]}
    sfp = kw_call(SubfieldParser, **{**kwargs, **kwargs_menc})
    itl = kw_call(bruma.iter_tl, mst_input, **kwargs, encoding=mst_encoding)
    for tl_decoded in itl:
        record = tl2record(tl_decoded, sfp, mode)
        ujson.dump(
            record, jsonl_output,
            ensure_ascii=ensure_ascii,
            escape_forward_slashes=False,
        )
        jsonl_output.write("\n")
        jsonl_output.flush()


@main.command()
@apply_decorators(*[op for op in mst_options if op.args[0] != "default_shift"])
@mst_ibp_option
@apply_decorators(*metadata_filtering_options)
@jsonl_mode_option
@apply_decorators(*subfield_options)
@utf8_fix_option
@file_arg_enc_option("mst", "rb", mst.DEFAULT_MST_ENCODING)
@file_arg_enc_option("jsonl", "w", DEFAULT_JSONL_ENCODING)
def mst2jsonl(mst_input, jsonl_output, mst_encoding, mode, utf8_fix, **kwargs):
    """ISIS/FFI Master File Format to JSON Lines."""
    ensure_ascii = jsonl_output.encoding.lower() == "ascii"
    mst_sc = kw_call(mst.StructCreator, **kwargs)
    sfp = kw_call(SubfieldParser, **kwargs)
    decode = utf8_fix_nest_decode if utf8_fix else nest_decode
    for tl in kw_call(mst_sc.iter_raw_tl, mst_input, **kwargs):
        record = tl2record(tl, sfp, mode)
        ujson.dump(
            decode(record, encoding=mst_encoding),
            jsonl_output,
            ensure_ascii=ensure_ascii,
            escape_forward_slashes=False,
        )
        jsonl_output.write("\n")
        jsonl_output.flush()


@main.command()
@apply_decorators(*mst_options)
@jsonl_mode_option
@apply_decorators(*subfield_options)
@subfield_unparse_check_option
@file_arg_enc_option("jsonl", "r", DEFAULT_JSONL_ENCODING)
@file_arg_enc_option("mst", OUTPUT_PATH, mst.DEFAULT_MST_ENCODING)
def jsonl2mst(jsonl_input, mst_output, mst_encoding, mode, **kwargs):
    """JSON Lines to ISIS/FFI Master File Format."""
    sfp = kw_call(SubfieldParser, **kwargs, check=kwargs["sfcheck"])
    mst_sc = kw_call(mst.StructCreator, **kwargs)
    def generate_records():
        for line in jsonl_input:
            record = nest_encode(ujson.loads(line), encoding=mst_encoding)
            tl = record2tl(record, sfp, mode)
            yield mst.tl2con(tl)
    with open(mst_output, "wb") as mst_file:
        mst_sc.build_stream(generate_records(), mst_file)


@main.command()
@apply_decorators(*iso_options)
@apply_decorators(*metadata_filtering_options)
@jsonl_mode_option
@apply_decorators(*subfield_options)
@utf8_fix_option
@file_arg_enc_option("iso", "rb", iso.DEFAULT_ISO_ENCODING)
@file_arg_enc_option("jsonl", "w", DEFAULT_JSONL_ENCODING)
def iso2jsonl(iso_input, jsonl_output, iso_encoding, mode, utf8_fix, **kwargs):
    """ISO2709 to JSON Lines."""
    ensure_ascii = jsonl_output.encoding.lower() == "ascii"
    kwargs["record_struct"] = kw_call(iso.create_record_struct, **kwargs)
    sfp = kw_call(SubfieldParser, **kwargs)
    decode = utf8_fix_nest_decode if utf8_fix else nest_decode
    for tl in kw_call(iso.iter_raw_tl, iso_input, **kwargs):
        record = tl2record(tl, sfp, mode)
        ujson.dump(
            decode(record, encoding=iso_encoding),
            jsonl_output,
            ensure_ascii=ensure_ascii,
            escape_forward_slashes=False,
        )
        jsonl_output.write("\n")
        jsonl_output.flush()


@main.command()
@apply_decorators(*iso_options)
@jsonl_mode_option
@apply_decorators(*subfield_options)
@subfield_unparse_check_option
@file_arg_enc_option("jsonl", "r", DEFAULT_JSONL_ENCODING)
@file_arg_enc_option("iso", "wb", iso.DEFAULT_ISO_ENCODING)
def jsonl2iso(jsonl_input, iso_output, iso_encoding, mode, **kwargs):
    """JSON Lines to ISO2709."""
    record_struct = kw_call(iso.create_record_struct, **kwargs)
    sfp = kw_call(SubfieldParser, **kwargs, check=kwargs["sfcheck"])
    for line in jsonl_input:
        record = nest_encode(ujson.loads(line), encoding=iso_encoding)
        tl = record2tl(record, sfp, mode)
        iso_bytes = iso.tl2bytes(tl, record_struct=record_struct)
        iso_output.write(iso_bytes)
        iso_output.flush()


if __name__ == "__main__":
    main()



from contextlib import contextmanager
from functools import wraps
from multiprocessing import Process, Pipe


@contextmanager
def jvm(domains=(), classpath=None, use_str=True):
    """JVM context using JPype.

    This must not be called more than once in a single process!
    """
    import jpype
    import jpype.imports
    jpype.startJVM(classpath=classpath, convertStrings=use_str)
    try:
        for domain in domains:
            jpype.imports.registerDomain(domain)
        yield
    finally:
        jpype.shutdownJVM()


def generator_blocking_process(func):  # noqa: C
    """Decorator to run a generator in another [blocking] process."""
    def run(func, conn_main, conn_proc, args, **kwargs):
        conn_main.close()
        try:
            conn_proc.recv()
        except EOFError:  # Pipe closed before asking the first entry
            pass
        else:
            for value in func(*args, **kwargs):
                try:
                    conn_proc.send(value)
                    conn_proc.recv()
                except EOFError:  # Broken pipe in the main process
                    break
        finally:
            conn_proc.close()

    @wraps(func)
    def wrapper(*args, **kwargs):
        conn_main, conn_proc = Pipe(duplex=True)
        proc = Process(
            target=run,
            args=[func, conn_main, conn_proc, args],
            kwargs=kwargs,
        )
        proc.start()
        try:
            conn_proc.close()
            while True:
                conn_main.send(None)  # Ask the next entry
                yield conn_main.recv()
        except EOFError:  # No more entries
            pass
        finally:
            conn_main.close()
            proc.join()

    return wrapper



from ioisis.java import generator_blocking_process, jvm


def test_jvm_get_datetime_multiple_calls():
    @generator_blocking_process
    def dt_gen(inc_days):
        with jvm():
            from java.time import LocalDateTime
            from java.time.format import DateTimeFormatter
            dt = LocalDateTime.of(2019, 11, 26, 16, 30, 0)
            dtfmt = DateTimeFormatter.ofPattern("dd/MM/yyyy HH:mm:ss")
            for inc in range(1, inc_days + 1):
                yield dt.plusDays(inc).format(dtfmt)

    # It should work more than once
    assert list(dt_gen(1)) == ["27/11/2019 16:30:00"]
    assert list(dt_gen(2)) == ["27/11/2019 16:30:00", "28/11/2019 16:30:00"]
    assert list(dt_gen(0)) == []



"""Model for the ISIS ISO2709-based file format.

This file format specification can be found at:

https://wiki.bireme.org/pt/img_auth.php/5/5f/2709BR.pdf
"""
from collections import defaultdict
from itertools import accumulate

from construct import Array, Bytes, Check, Computed, \
                      Const, Container, Default, ExprAdapter, \
                      FocusedSeq, Prefixed, RawCopy, \
                      Rebuild, Select, Struct, Terminated, this

from .ccons import IntInASCII, LineSplitRestreamed, \
                   DEFAULT_LINE_LEN, DEFAULT_NEWLINE
from .streamutils import should_be_file, TightBufferReadOnlyBytesStreamWrapper


DEFAULT_FIELD_TERMINATOR = b"#"
DEFAULT_RECORD_TERMINATOR = b"#"
DEFAULT_ISO_ENCODING = "cp1252"

TOTAL_LEN_LEN = 5
LEADER_LEN = TOTAL_LEN_LEN + 19
TAG_LEN = 3
DEFAULT_LEN_LEN = 4
DEFAULT_POS_LEN = 5
DEFAULT_CUSTOM_LEN = 0


def create_record_struct(
    field_terminator=DEFAULT_FIELD_TERMINATOR,
    record_terminator=DEFAULT_RECORD_TERMINATOR,
    line_len=DEFAULT_LINE_LEN,
    newline=DEFAULT_NEWLINE,
):
    """Create a construct parser/builder for a whole record object."""
    ft_len = len(field_terminator)
    prefixless = Struct(
        # Build time pre-computed information
        "_build_len_list" / Computed(
            lambda this: None if "fields" not in this else
                [len(field) + ft_len for field in this.fields]
        ),
        "_build_pos_list" / Computed(
            lambda this: None if "fields" not in this else
                list(accumulate([0] + this._build_len_list))
        ),
        "_build_dir_len" / Computed(
            lambda this: None if "fields" not in this else
                len(this.fields) * (
                    TAG_LEN
                    + this.get("len_len", DEFAULT_LEN_LEN)
                    + this.get("pos_len", DEFAULT_POS_LEN)
                    + this.get("custom_len", DEFAULT_CUSTOM_LEN)
                )
        ),

        # Record leader/header (apart from the leading total_len)
        "status" / Default(Bytes(1), b"0"),
        "type" / Default(Bytes(1), b"0"),
        "custom_2" / Default(Bytes(2), b"00"),
        "coding" / Default(Bytes(1), b"0"),
        "indicator_count" / Default(IntInASCII(Bytes(1)), 0),
        "identifier_len" / Default(IntInASCII(Bytes(1)), 0),
        "base_addr" / Rebuild(IntInASCII(Bytes(5)),
                              LEADER_LEN + this._build_dir_len
                                         + ft_len),
        "custom_3" / Default(Bytes(3), b"000"),

        # Directory entry map (trailing part of the leader)
        "len_len" / Default(IntInASCII(Bytes(1)), DEFAULT_LEN_LEN),
        "pos_len" / Default(IntInASCII(Bytes(1)), DEFAULT_POS_LEN),
        "custom_len" / Default(IntInASCII(Bytes(1)),
                               DEFAULT_CUSTOM_LEN),
        "reserved" / Default(Bytes(1), b"0"),

        # The ISO leader/header doesn't have the number of fields,
        # but it can be found from the base address
        "num_fields" / Computed(
            (this.base_addr - LEADER_LEN - ft_len) //
            (TAG_LEN + this.len_len + this.pos_len + this.custom_len)
        ),
        Check(lambda this:
            "fields" not in this or this.num_fields == len(this.fields)
        ),

        # Directory
        "dir" / Struct(
            "tag" / Bytes(TAG_LEN),
            "len" / Rebuild(IntInASCII(Bytes(this._.len_len)),
                            lambda this: this._._build_len_list[this._index]),
            "pos" / Rebuild(IntInASCII(Bytes(this._.pos_len)),
                            lambda this: this._._build_pos_list[this._index]),
            "custom" / Default(Bytes(this._.custom_len),
                               b"0" * this._.custom_len),
        )[this.num_fields],
        Check(lambda this: this.num_fields == 0 or (
            this.dir[0].pos == 0 and
            all(
                this.dir[idx + 1].pos == entry.pos + entry.len
                for idx, entry in enumerate(this.dir[:-1])
            )
        )),
        Const(field_terminator),
        Check(lambda this: this._io.tell() + TOTAL_LEN_LEN == this.base_addr),

        # Field data
        "fields" / Array(
            this.num_fields,
            FocusedSeq(
                "value",
                "value" / Bytes(
                    lambda this: this._.dir[this._index].len - ft_len
                ),
                Const(field_terminator),
            ),
        ),

        # There should be no more data belonging to this record
        Const(record_terminator),
    )

    # This includes (and checks) the total_len prefix
    result = ExprAdapter(
        RawCopy(Prefixed(
            lengthfield=IntInASCII(Bytes(TOTAL_LEN_LEN)),
            subcon=prefixless,
            includelength=True,
        )),
        lambda obj, ctx: Container(total_len=obj.length, **obj.value),
        lambda obj, ctx: {"value": obj},
    )

    if line_len is None or line_len == 0:
        return result
    return LineSplitRestreamed(result, line_len=line_len, newline=newline)


DEFAULT_RECORD_STRUCT = create_record_struct()


@should_be_file("iso_file")
def iter_con(iso_file, record_struct=DEFAULT_RECORD_STRUCT):
    """Generator of records as parsed construct objects."""
    alt_struct = Select(record_struct, Terminated)
    while True:
        stream_reader = TightBufferReadOnlyBytesStreamWrapper(iso_file)
        con = alt_struct.parse_stream(stream_reader)
        if con is None:  # No more records
            return
        yield con


def iter_records(iso_file, encoding=DEFAULT_ISO_ENCODING, **kwargs):
    """Generator of records as dictionaries."""
    for con in iter_con(iso_file, **kwargs):
        yield con2dict(con, encoding=encoding)


def con_pairs(con):
    """Generator of raw ``(tag, field)`` pairs of ``bytes`` objects."""
    for dir_entry, field_value in zip(con.dir, con.fields):
        yield dir_entry.tag.lstrip(b"0") or b"0", field_value


def iter_raw_tl(iso_file, *, only_active=True, prepend_status=False,
                record_struct=DEFAULT_RECORD_STRUCT):
    for con in iter_con(iso_file, record_struct=record_struct):
        if only_active and con.status != b"0":
            continue
        result = []
        if prepend_status:
            result.append((b"status", b"%d" % con.status))
        result.extend(con_pairs(con))
        yield result


def iter_tl(iso_file, encoding=DEFAULT_ISO_ENCODING, **kwargs):
    for tl in iter_raw_tl(iso_file, **kwargs):
        yield [(tag.decode("ascii"), field.decode(encoding))
               for tag, field in tl]


def con2dict(con, encoding=DEFAULT_ISO_ENCODING):
    """Parsed construct object to dictionary record converter."""
    result = defaultdict(list)
    for tag_value, field_value in con_pairs(con):
        result[tag_value.decode("ascii")].append(field_value.decode(encoding))
    return result


def dict2bytes(
    data,
    encoding=DEFAULT_ISO_ENCODING,
    record_struct=DEFAULT_RECORD_STRUCT,
):
    """Encode/build the raw ISO string from a single dict record."""
    record_dict = {
        "dir": [],
        "fields": [],
    }
    for k, values in data.items():
        for v in values:
            record_dict["dir"].append({"tag": k.encode("ascii").zfill(3)})
            record_dict["fields"].append(v.encode(encoding))
    return record_struct.build(record_dict)


def tl2bytes(tl, record_struct=DEFAULT_RECORD_STRUCT):
    """Encode/build the raw ISO string from a single tidy list record."""
    container = {
        "dir": [],
        "fields": [],
    }
    for k, v in tl:
        container["dir"].append({"tag": k.zfill(3)})
        container["fields"].append(v)
    return record_struct.build(container)



from click.testing import CliRunner

from ioisis.__main__ import iso2jsonl, jsonl2iso


simple_example_jsonl = (
    b'{"1":["a"]}\n'
    b'{"10":["test","one"],"11":["two"]}\n'
    b'{}\n'
    b'{"1":["x"],"10":["y","z"],"100":["aa","bbb","cccc"]}\n'
)

simple_example_iso = (
    b"000400000000000370004500001000200000#a##\n"
    b"000750000000000610004500"
        b"010000500000010000400005011000400009#test#one#two##\n"
    b"000260000000000250004500##\n"
    b"001160000000000970004500"
        b"00100020000001000020000201000020000410000030000610000040\n"
    b"0009100000500013#x#y#z#aa#bbb#cccc##\n"
)


def test_jsonl2iso_simple_example_on_standard_streams():
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(jsonl2iso, input=simple_example_jsonl)
    assert result.exit_code == 0
    assert result.stdout_bytes == simple_example_iso
    assert result.stderr_bytes == b""


def test_iso2jsonl_simple_example_on_standard_streams():
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(iso2jsonl, input=simple_example_iso)
    assert result.exit_code == 0
    assert result.stdout_bytes == simple_example_jsonl
    assert result.stderr_bytes == b""



"""Model for the ISIS master file format.

Part of the ISIS-mode packed unshifted file format specification
can be found at:

https://unesdoc.unesco.org/ark:/48223/pf0000211280

The unpacked, FFI, shifted and other customized file formats
were discovered based on the analysis of actual MST files
as well as the source code of CISIS and Bruma.
"""
from binascii import b2a_hex
from itertools import accumulate

from construct import Array, BitsInteger, BitStruct, \
                      Byte, Bytes, ByteSwapped, \
                      Check, CheckError, Computed, Const, \
                      Default, ExprAdapter, Flag, FocusedSeq, \
                      Int16sb, Int16sl, Int16ub, Int16ul, \
                      Int32sb, Int32sl, Int32ub, Int32ul, \
                      Padded, Padding, Rebuild, Select, SelectError, Struct, \
                      Tell, Terminated, Union

from .ccons import DictSegSeq, Unnest


DEFAULT_MST_ENCODING = "cp1252"
DEFAULT_ENDIANNESS = "little"
DEFAULT_FORMAT = "isis"
DEFAULT_LOCKABLE = True
DEFAULT_SHIFT = 6
DEFAULT_SHIFT4IS3 = False
DEFAULT_MIN_MODULUS = 2
DEFAULT_PACKED = False
DEFAULT_FILLER = b"\x00"
DEFAULT_RECORD_FILLER = b" "
DEFAULT_CONTROL_LEN = 64
DEFAULT_IBP = "check"


def con_pairs(con):
    """Generator of raw ``(tag, field)`` pairs of ``bytes`` objects."""
    for dir_entry, field_value in zip(con.dir, con.fields):
        yield b"%d" % dir_entry.tag, field_value


def tl2con(tl):
    """Create a record dict that can be used for MST building
    from a single tidy list record."""
    container = {
        "dir": [],
        "fields": [],
    }
    for k, v in tl:
        if k == b"mfn":
            container["mfn"] = int(v)
        else:
            container["dir"].append({"tag": int(k)})
            container["fields"].append(v)
    return container


def pad_size(modulus, size):
    """Calculate the padding size for the given size in a modulus grid."""
    return (modulus * (size // modulus + 1) - size) % modulus


def never_split_pad_size(address, leader_len):
    """Calculate the padding until the end of block for the next record."""
    offset = address & 0x1ff
    return 0 if offset + leader_len - 4 <= 512 else 512 - offset


class StructCreator:
    """
    Creator of MST record structs.

    Parameters
    ----------
    endianness : str
        Byte order endianness for all 16/32 bits integer numbers,
        might be "big" or "little" (a.k.a. swapped).
    format : str
        Format mode, a string that should either be "isis" or "ffi".
        The FFI is a format mode that supports bigger records,
        with sizes up to 2_147_483_647 instead of 32767
        (or twice that plus one if lockable is ``False``).
        To do that,
        the records in the FFI format mode have 4 bytes instead of 2
        in two leader fields (``mfrl`` and ``base_addr``)
        and two directory fields (``pos`` and ``len``).
    lockable : bool
        Multi-user locking, enabling three locking mechanisms:
        ``delock_count`` (Data entry lock),
        ``ewlock`` (Exclusive write lock), and
        ``rlock`` (Record lock).
        The latter is stored in the ``mfrl`` sign of a record,
        the former ones are simply an interpretation
        of respectively the ``mfcxx2`` and ``mfcxx3`` fields
        from the control record,
        where the ``delock_count`` is a counter
        of how many records has the ``rlock``,
        and ``ewlock`` is a file-level lock
        for a single application to block other ones from writing,
        e.g. during a backup/restore operation.
        Locked records might have a 9999 tag,
        which would include the timestamp and user who locked it.
    default_shift : int
        Default XRF shifting for building,
        it tells how many bits should be shifted in the XRF pointers.
        On parsing, this is obtained from the ``mstxl`` field
        (the most significant byte of the legacy MSTYPE field)
        in the control record.
    shift4is3 : bool
        Flag for the legacy shifting interpretation
        where a shift of 4 should be regarded as 3.
    min_modulus : int
        Smallest padding "modulus" for MST records alignment.
        The modulus is ``2 ** shift``,
        unless this number is higher than that.
        It must be a power of 2.
        The standard ISIS defines it as 2.
    packed : bool
        Flag for packing the leader and directory fields
        with no filler/padding/slack bytes for DWORD alignment.
        It disables the 4-bytes alignment
        of all 4-bytes integer fields.
        If ``False``, it adds a WORD-sized (16 bits) padding
        for DWORD-sized (32 bits) alignment
        in all fields that requires it, that is,
        this padding filler appears after the ``mfrl`` in ISIS,
        and after ``base_addr`` and ``tag`` in FFI.
        The origin of these fillers for data structure alignment
        is historical, coming from the usage of raw structs in C
        without neither ``#pragma pack(2)`` or ``-fpack-struct=2``
        or anything else that enforces the same packing,
        and the fact that CISIS stores data in master files
        in the same way they are stored in memory,
        reflecting any implementation detail from the ABI,
        the C/C++ compiler and the architecture.
        Most C/C++ compilers, including GCC,
        standardized the data alignment to always have
        ``int``/``long`` (4 bytes fields) 4-byte aligned
        in 32 bits architectures,
        so this flag should almost always be ``False``
        for files created with CISIS,
        although it's not part of the original ISIS format.
        The file format has nothing specific to the operating system,
        but ``packed=False'' might also be called "Linux version",
        and ``packed=True'' the "Windows version"
        in some Bruma and CISIS code and documentation,
        because CISIS used to be compiled with ``-fpack-struct=1``
        in the latter as a consequence of a custom IDE configuration,
        something that had never been applied
        when compiling it on Linux.
        Bruma names this using the number of slack bytes,
        where its *align2* means ``packed=False''
        and its *align0* means ``packed=True'',
        although the actual alignment for each case
        is respectively 4-bytes and 2-bytes.
    filler : bytes,
        Default filler character,
        used to replace ``None`` in the remaining filler arguments.
    control_filler : bytes or None
        Filler character for the trailing bytes of the control record.
        In some obscure cases of this might be ``"\\xff"``,
        as CISIS describes for Unisys.
    slack_filler : bytes or None
        Padding filler character
        to appear between some leader and directory fields
        when ``packed=False''.
    block_filler : bytes or None
        Filler character for trailing recordless bytes of a block.
    record_filler : bytes or None
        Filler character for the trailing record data
        that doesn't belong to any field,
        but appears due to the record alignment.
    control_len : int
        Control record length,
        it should be at least ``max(32, 2 ** shift)``.
    ibp : str
        Invalid block padding content/size acceptance mode.
        There are 3 modes:
        "check", to raise an exception when such invalid data appears;
        "ignore", to skip these bytes;
        and "store", to add an "ibp" field
        with the found trailing invalid data in hex.
    """
    def __init__(
        self,
        endianness=DEFAULT_ENDIANNESS,
        format=DEFAULT_FORMAT,
        lockable=DEFAULT_LOCKABLE,
        default_shift=DEFAULT_SHIFT,
        shift4is3=DEFAULT_SHIFT4IS3,
        min_modulus=DEFAULT_MIN_MODULUS,
        packed=DEFAULT_PACKED,
        filler=DEFAULT_FILLER,
        control_filler=None,
        slack_filler=None,
        block_filler=None,
        record_filler=DEFAULT_RECORD_FILLER,
        control_len=DEFAULT_CONTROL_LEN,
        ibp=DEFAULT_IBP,
    ):
        # Get the actual value for every filler
        self.control_filler = \
            filler if control_filler is None else control_filler
        self.slack_filler = \
            filler if slack_filler is None else slack_filler
        self.block_filler = \
            filler if block_filler is None else block_filler
        self.record_filler = \
            filler if record_filler is None else record_filler

        # Validation for inputs that wouldn't break somewhere else
        if endianness not in ["big", "little"]:
            raise ValueError("Invalid endianness")
        if format not in ["isis", "ffi"]:
            raise ValueError("Invalid format mode")
        if ibp not in ["check", "ignore", "store"]:
            raise ValueError("Invalid ibp mode")
        min_modulus_rounded_to_powerof2 = 1 << (min_modulus.bit_length() - 1)
        if min_modulus <= 0 or min_modulus != min_modulus_rounded_to_powerof2:
            raise ValueError("Value of min_modulus must be "
                             "a positive power of 2")
        if control_len % min_modulus != 0:
            raise ValueError("Control record length "
                             "isn't a multiple of min_modulus")
        if any(len(f) != 1 for f in [filler,
                                     self.control_filler, self.slack_filler,
                                     self.block_filler, self.record_filler]):
            raise ValueError("All filler patterns lengths must be 1")

        # Store the inputs
        self.endianness = endianness
        self.format = format
        self.lockable = lockable
        self.default_shift = default_shift
        self.shift4is3 = shift4is3
        self.min_modulus = min_modulus
        self.packed = packed
        self.control_len = control_len
        self.ibp = ibp

    def create_control_record_struct(self):
        little_endian = self.endianness == "little"
        min_mod = self.min_modulus
        control_len = self.control_len

        # Data types for integer values
        Int16s, Int16u, Int32s, Int32u = {
            "big":    (Int16sb, Int16ub, Int32sb, Int32ub),
            "little": (Int16sl, Int16ul, Int32sl, Int32ul),
        }[self.endianness]

        # The 16 bytes that used to have the legacy MFTYPE field
        # has both the actual MFTYPE and the MSTXL,
        # and their order is affected by the endianness
        mftype = "mftype" / Default(Byte, 0)
        mstxl = "mstxl" / Rebuild(
            Byte,
            lambda this:
                this.mstxl if "mstxl" in this and this.mstxl is not None else
                this.get(
                    "shift",
                    this.modulus.bit_length() - 1
                        if "modulus" in this else self.default_shift,
                ),
        )

        # MFCXX2 and MFCXX3 are used for locking the file on updating
        if self.lockable:
            mfcxx2 = "_mfcxx2" / Union(0,
                "mfcxx2" / Default(Int32s, 0),
                "delock_count" / Int32u,  # Data entry lock
            )
            mfcxx3 = "_mfcxx3" / Union(0,
                "mfcxx3" / Default(Int32s, 0),
                "ewlock" / ExprAdapter(Int32u,  # Exclusive write lock
                                       lambda obj, ctx: bool(obj),
                                       lambda obj, ctx: int(obj)),
            )
        else:
            mfcxx2 = "mfcxx2" / Default(Int32s, 0)
            mfcxx3 = "mfcxx3" / Default(Int32s, 0)

        # Legacy shift replacement
        if self.shift4is3:
            shift = "shift" / Computed(lambda this:
                                       3 if this.mstxl == 4 else this.mstxl)
        else:
            shift = "shift" / Computed(lambda this: this.mstxl)

        # The control record struct for all cases
        nested_unpadded_struct = Struct(
            # First 4 fields, including information about the whole file
            "mfn" / Default(Int32s, 0),  # CTLMFN
            Check(lambda this: this.mfn == 0),
            "next_mfn" / Default(Int32s, 1),  # NXTMFN
            "next_block" / Default(Int32s, 1),  # NXTMFB
            "next_offset" / Default(Int16u, self.control_len),  # NXTMFP
            *([mftype, mstxl] if little_endian else [mstxl, mftype]),

            # Get the MST (modulus) and XRF (shift) alignment/shift values
            # from the MSTXL field, fixing the "legacy" replacements
            shift,
            "modulus" / Computed(lambda this: max(min_mod, 1 << this.shift)),
            Check(lambda this: control_len % this.modulus == 0),

            # Fields used for "statistics during backup/restore",
            # where the last two fields are also used for multi-user locking
            "reccnt" / Default(Int32s, 0),
            "mfcxx1" / Default(Int32s, 0),
            mfcxx2,
            mfcxx3,
        )
        return Padded(
            length=control_len,
            subcon=Unnest(["_mfcxx2", "_mfcxx3"], nested_unpadded_struct)
                if self.lockable else nested_unpadded_struct,
            pattern=self.control_filler,
        )

    def create_record_struct(self, control_record):
        # Pre-computed lengths and flags
        ffi = self.format == "ffi"
        slacked = not self.packed
        leader_len = 18 + 4 * ffi + 2 * slacked
        dir_entry_len = 6 + 4 * ffi + 2 * (ffi & slacked)
        min_mod = max(self.min_modulus, 1 << self.default_shift)
        inf = float("inf")

        # Data types for integer values (strings are always big endian)
        Int16s, Int16u, Int32s, Int32u = {
            "big":    (Int16sb, Int16ub, Int32sb, Int32ub),
            "little": (Int16sl, Int16ul, Int32sl, Int32ul),
        }[self.endianness]

        # For the leader and directory structure of the records,
        # some sizes are configured based on the ffi and lockable flags
        # (MRFL is the record length)
        if self.lockable:
            mfrl_fields = [
                "mfrl" / Rebuild(
                    Int32s if ffi else Int16s,
                    lambda this:
                        -this._build_total_len
                        if this.get("rlock", False) else
                        this._build_total_len
                ),
                "total_len" / Computed(lambda this: abs(this.mfrl)),
                "rlock" / Computed(lambda this: this.mfrl < 0),
            ]
        else:
            mfrl_fields = [
                "mfrl" / Rebuild(Int32u if ffi else Int16u,
                                 lambda this: this._build_total_len),
                "total_len" / Computed(lambda this: this.mfrl),
            ]

        return Struct(
            # Block alignment ("never splits" the leader unless BASE fits)
            "_before_start" / Tell,
            Padding(
                lambda this:
                    never_split_pad_size(this._before_start, leader_len),
                self.block_filler,
            ),

            "_start" / Tell,

            # Build time pre-computed information
            "_build_len_list" / Computed(
                lambda this: None if "fields" not in this else
                    [len(field) for field in this.fields]
            ),
            "_build_pos_list" / Computed(
                lambda this: None if "fields" not in this else
                    list(accumulate([0] + this._build_len_list))
            ),
            "_build_total_len_padless" / Computed(
                lambda this: None if "fields" not in this else
                    leader_len +
                    dir_entry_len * len(this.fields) +
                    this._build_pos_list[-1]  # Fields length
            ),
            "_build_total_len" / Computed(
                lambda this: None if "fields" not in this else
                    this._build_total_len_padless +
                    pad_size(control_record.get("modulus", min_mod),
                             this._build_total_len_padless)
            ),

            # Record leader/header
            "mfn" / Int32s,  # Master file number
            Check(lambda this: this.mfn != 0),  # Not a control record
            Check(lambda this: control_record.get("next_mfn", inf) > this.mfn),
            *mfrl_fields,

            Check(lambda this:
                  this.total_len % control_record.get("modulus", min_mod)
                  == 0),
            *([Const(self.slack_filler * 2)] if slacked and not ffi else []),
            "old_block" / Default(Int32s, 0),  # MFBWB backward pointer block
            "old_offset" / Default(Int16u, 0),  # MFBWP backward pointer offset
            *([Const(self.slack_filler * 2)] if slacked and ffi else []),
            "base_addr" / Rebuild(
                Int32u if ffi else Int16u,
                lambda this: leader_len + dir_entry_len * len(this.fields),
            ),
            "num_fields" / Rebuild(Int16u,  # NVF
                                   lambda this: len(this.fields)),
            "status" / Default(Int16u, 0),  # Active is 0, deleted is 1

            # Directory
            "dir" / Struct(
                "tag" / Int16u,
                *([Const(self.slack_filler * 2)] if slacked and ffi else []),
                "pos" / Rebuild(
                    Int32u if ffi else Int16u,
                    lambda this: this._._build_pos_list[this._index],
                ),
                "len" / Rebuild(
                    Int32u if ffi else Int16u,
                    lambda this: this._._build_len_list[this._index],
                ),
            )[lambda this: this.num_fields],
            "_before_fields" / Tell,
            Check(lambda this:
                  this.base_addr == this._before_fields - this._start),

            # Field data
            "fields" / Array(
                lambda this: this.num_fields,
                "value" / Bytes(lambda this: this.dir[this._index].len),
            ),

            # Record alignment (implicitly checks the length stored in MFRL)
            "_after_fields" / Tell,
            Padding(
                lambda this:
                    this.total_len - (this._after_fields - this._start),
                self.record_filler,
            ),
        )

    def create_ending_struct(self):
        return FocusedSeq(
            "empty",
            "_before_start" / Tell,
            Padding(
                lambda this: 512 - (this._before_start & 0x1ff),
                self.block_filler,
            ),
            "empty" / Terminated,
        )

    def create_xrf_struct(self, control_record):
        shift = control_record.shift
        Int32s, BitStructWrapper = {
            "big": (Int32sb, lambda x: x),
            "little": (Int32sl, ByteSwapped),
        }[self.endianness]
        return FocusedSeq(
            "data",
            "data" / DictSegSeq(
                idx_field=Int32s,
                subcon=BitStructWrapper(BitStruct(
                    "block" / Default(BitsInteger(21 + shift, signed=True), 0),
                    "is_new" / Default(Flag, False),
                    "is_updated" / Default(Flag, False),
                    "offset" / Default(ExprAdapter(
                        BitsInteger(9 - shift),
                        lambda obj, context: obj << shift,
                        lambda obj, context: obj >> shift,
                    ), 0),
                )),
                block_size=127,  # Not counting the index field
                empty_item={},
                check_nonempty=lambda item: any([
                    item.block, item.offset, item.is_new, item.is_updated,
                ]),
            ),
            Terminated,
        )

    def iter_con(self, mst_stream, yield_control_record=False): # noqa: C
        """Generator of records as parsed construct.Container objects.

        Parameters
        ----------
        mst_stream : file-like
            A seekable open stream of bytes in its initial position,
            where the MST binary contents should be read from.
        yield_control_record : bool
            Control if the first generated construct.Container object
            should be the control record,
            or if it should skip it to generate just the records.
        """
        ending_struct = self.create_ending_struct()
        control_record_struct = self.create_control_record_struct()
        control_record = control_record_struct.parse_stream(mst_stream)
        record_struct = self.create_record_struct(control_record)
        leader_len = 18 + 4 * (self.format == "ffi") + 2 * (not self.packed)
        rec_or_end_struct = Select(record_struct, ending_struct)

        last_tell = 0
        def record_ibp_gen():
            nonlocal last_tell
            ibps = []
            while True:
                try:
                    record = rec_or_end_struct.parse_stream(mst_stream)
                except SelectError:
                    if self.ibp == "check":
                        raise
                    elif self.ibp == "store":
                        ibps.append(mst_stream.read(control_record.modulus))
                    continue
                if ibps:
                    yield {"ibp": b2a_hex(b"".join(ibps))}
                    ibps.clear()
                if record is None:  # No more records
                    break
                last_tell = mst_stream.tell()
                yield record

        if yield_control_record:
            prev = control_record
        else:
            prev = None
        for record in record_ibp_gen():
            if record and "ibp" in record:
                prev.update(record)
                yield prev
                prev = None
            else:
                if prev:
                    yield prev
                prev = record
        if prev:
            yield prev

        next_addr = last_tell + never_split_pad_size(last_tell, leader_len)
        next_block = 1 + (next_addr >> 9)
        next_offset = 1 + (next_addr & 0x1ff)
        if control_record.next_block != next_block:
            raise CheckError("Invalid next_block")
        if control_record.next_offset != next_offset:
            raise CheckError("Invalid next_offset")

    def iter_raw_tl(self, mst_stream, *,
                    only_active=True, prepend_mfn=False, prepend_status=False,
    ):
        for con in self.iter_con(mst_stream):
            if con.get("old_block", 0) != 0 or con.get("old_offset", 0) != 0:
                raise NotImplementedError("Pending master file reorganization")
            if only_active and con.status != 0:
                continue
            result = []
            if prepend_mfn:
                result.append((b"mfn", b"%d" % con.mfn))
            if prepend_status:
                result.append((b"status", b"%d" % con.status))
            result.extend(con_pairs(con))
            if "ibp" in con and self.ibp == "store":
                result.append((b"ibp", con["ibp"]))
            yield result

    def build_stream(self, records, mst_stream, control_record=None):
        """Build the MST binary data on the given seekable stream.

        Parameters
        ----------
        records : iterable of dict or construct.Container instances
            A preferably lazy iterable (generator)
            with the records to be stored, but the control_record,
            ordered by their MFN.
            If the records don't include the MFN number,
            a serial one will be included,
            resuming it from the biggest MFN previously included.
        mst_stream : file-like
            An empty, writeable and seekable open stream of bytes
            to store the MST binary contents.
        control_record : dict or construct.Container
            A custom control record object,
            allowing the storage of data in fields like reccnt.
        """
        ending_struct = self.create_ending_struct()

        control_record_struct = self.create_control_record_struct()
        if control_record is None:
            control_record = {}
        control_record_struct.build_stream(control_record, mst_stream)
        mst_stream.flush()

        record_struct = self.create_record_struct(control_record)
        leader_len = 18 + 4 * (self.format == "ffi") + 2 * (not self.packed)

        next_mfn = 1
        for record in records:
            if "mfn" in record:
                next_mfn = max(next_mfn, record["mfn"] + 1)
            else:
                record["mfn"] = next_mfn
                next_mfn += 1
            record_struct.build_stream(record, mst_stream)
            mst_stream.flush()
        last_tell = mst_stream.tell()
        ending_struct.build_stream(None, mst_stream)
        mst_stream.flush()

        next_addr = last_tell + never_split_pad_size(last_tell, leader_len)
        control_record["next_mfn"] = next_mfn
        control_record["next_block"] = 1 + (next_addr >> 9)
        control_record["next_offset"] = 1 + (next_addr & 0x1ff)

        end_tell = mst_stream.tell()
        mst_stream.seek(0)
        control_record_struct.build_stream(control_record, mst_stream)
        mst_stream.flush()
        mst_stream.seek(end_tell)



from collections import Counter, defaultdict
from itertools import cycle, zip_longest
import re


# The UTF-8 bytes (and number of bits to store a code point) are:
#
# - 0b0xxx_xxxx: single-byte sequence (ASCII)
# - 0b10xx_xxxx: continuation byte
# - 0b110x_xxxx: start byte of a 2-bytes sequence (11 CP bits)
# - 0b1110_xxxx: start byte of a 3-bytes sequence (16 CP bits)
# - 0b1111_0xxx: start byte of a 4-bytes sequence (21 CP bits)
#
# Code points of ASCII bytes should not be "overlong" UTF-8 sequences:
# 2-bytes sequences can encode unicode up to 11 code point bits,
# but we need to enforce it to have at least 8 code point bits,
# so the lowest 2-bytes valid sequence in UTF-8 is 0xc280 (U+0080).
# Likewise, the last unicode code point is U+10ffff,
# and a sequence should not represent a code point greater than that,
# so the highest 4-bytes UTF-8 is 0xf48fbfbf.
# Sequences should be stored with as fewer bytes as possible,
# so the smallest 3-bytes UTF-8 is 0xe0a080 (U+0800)
# and the smallest 4-bytes UTF-8 is 0xf0908080 (U+010000).
#
# Apart from that, the code points in the U+D800 to U+DFFF range
# aren't UTF-8 valid because they had been reserved
# for encoding data as surrogate pairs in UTF-16.
# It means that the range from 0xeda080 to 0xedbfbf is invalid.
#
# See also RFC3629 (https://tools.ietf.org/html/rfc3629)
# for an authoritative source of the information above.
UTF8_MB_REGEX = re.compile(  # Gets joined UTF-8 multi-byte sequences
    b"((?:[\\xc2-\\xdf][\\x80-\\xbf]"
    b" |   \\xe0       [\\xa0-\\xbf][\\x80-\\xbf]"
    b" |  [\\xe1-\\xec][\\x80-\\xbf][\\x80-\\xbf]"
    b" |   \\xed       [\\x80-\\x9f][\\x80-\\xbf]"
    b" |  [\\xee-\\xef][\\x80-\\xbf][\\x80-\\xbf]"
    b" |   \\xf0       [\\x90-\\xbf][\\x80-\\xbf][\\x80-\\xbf]"
    b" |  [\\xf1-\\xf3][\\x80-\\xbf][\\x80-\\xbf][\\x80-\\xbf]"
    b" |   \\xf4       [\\x80-\\x8f][\\x80-\\xbf][\\x80-\\xbf]"
    b" )+)",
    re.VERBOSE,
)

_EMPTY = object()


class SubfieldParser:
    """Generate subfield pairs from the given value on calling.

    Parameters
    ----------
    prefix : str or bytes
        Marker of the beginning of a new subfield.
        The type of this prefix should be the same
        of the fields to be parsed.
    length : int
        Subfield key length after its mark in the field.
    lower : bool
        Force keys to be in lowercase, making them case insensitive.
    first : str, bytes or None
        The key to be used for the leading value.
    empty : bool
        Keep pairs whose value is empty.
    number : bool
        Add a number suffix in all but the first
        of each distinct subfield key,
        grouping this indexing by the key
        and starting the numbering an implicit ``0``,
        so the subfield with a key that appeared for the second time
        should have a ``1'' suffix.
    zero : bool
        Also append the ``0`` in subfields keys,
        so that all keys should have a suffix
        (including the first/leading "keyless" pair).
        Has no effect if ``number`` is ``False``.
    check : bool
        Check data consistency on unparsing.
        See ``SubfieldParser.unparse`` for more information.

    Examples
    --------
    It can be used to convert the field to a dict:

    >>> gen_subfields = SubfieldParser(b"^", first=b"_")
    >>> dict(gen_subfields(b"data^ttext^len^tTrail"))
    {b'_': b'data', b't': b'text', b'l': b'en', b't1': b'Trail'}

    Or to a list, and the type (bytes/str) is defined by the prefix:

    >>> str_subfields = SubfieldParser("#F#", number=False)
    >>> list(str_subfields("data#F#ttext#F#len#F#tTrail"))
    [('', 'data'), ('t', 'text'), ('l', 'en'), ('t', 'Trail')]

    """
    def __init__(self, prefix, *, length=1, lower=False, first=None,
                 empty=False, number=True, zero=False, check=True):
        self.prefix = prefix
        self.length = length
        self.lower = lower
        self.empty = empty
        self.number = number
        self.zero = zero
        self.check = check

        escaped_prefix = re.escape(prefix)
        regex_str = b"(?:^|(?<=%s(.{%d})))((?:(?!%s.{%d}).)*)"
        if isinstance(prefix, str):
            regex_str = regex_str.decode("ascii")
            self.percent_d = "%d"
        else:
            self.percent_d = b"%d"
        regex_str %= (escaped_prefix, length) * 2
        self.subfields_regex = re.compile(regex_str, re.DOTALL)

        if first is None:
            self.first = regex_str[:0]  # Empty bytes or str
        elif lower:
            self.first = first.lower()
        else:
            self.first = first

        if self.number and self.zero:
            self.fz = self.first + (self.percent_d % 0)
        else:
            self.fz = self.first

    def __call__(self, field):
        """Generate (key, value) pairs for each subfields in a field."""
        key_count = Counter()
        for key, value in self.subfields_regex.findall(field):
            if self.empty or value:
                if not key:  # PyPy: empty key is always str, not bytes
                    key = self.first
                elif self.lower:
                    key = key.lower()
                if self.number:
                    suffix_int = key_count[key]
                    key_count[key] += 1
                    if self.zero or suffix_int:
                        key += self.percent_d % suffix_int
                yield key, value

    def unparse(self, *subfields, check=_EMPTY):
        """Build the field from the ordered subfield pairs.

        Parameters
        ----------
        *subfields
            Subfields as ``(key, value)`` pairs (tuples)
            of ``bytes`` or str``.
        check : bool
            Force checking if this SubfieldParser
            would generate exactly the same subfields from the result.
            That won't happen if the input can't be created
            by this SubfieldParser instance,
            so it's a way to check the number suffixes,
            empty subfields, subfields keys in upper case,
            and invalid contents like a subfield inside another.
        """
        blocks = []
        if subfields and ((subfields[0][0] == self.fz) or
                          (self.lower and subfields[0][0].lower() == self.fz)):
            blocks.append(subfields[0][1])
            remaining = subfields[1:]
        else:
            remaining = subfields

        for key, value in remaining:
            if self.empty or value:
                if len(key) < self.length:
                    raise ValueError(f"Incomplete key data {key!r}")
                if self.lower:
                    key = key.lower()
                blocks.append(self.prefix + key[:self.length] + value)

        result = self.prefix[:0].join(blocks)
        has_to_check = self.check if check is _EMPTY else check
        if has_to_check:
            self._parse_check(result, *subfields)
        return result

    def _parse_check(self, field, *subfields):
        """Check if the subfields are the parsed field."""
        parsed = self(field)
        pairs_of_pairs = zip_longest(parsed, subfields, fillvalue=(None, None))
        for idx, ((kp, vp), (ks, vs)) in enumerate(pairs_of_pairs):
            if ks != kp:
                raise ValueError(f"Invalid subfield[{idx}] key {ks!r}")
            if vs != vp:
                raise ValueError(f"Invalid subfield[{idx}] value {vs!r}")


def tl2record(tl, sfp=None, mode="field"):
    """Converter of a record from a tidy list to a dictionary."""
    if mode in "field":
        items = tl
    elif mode == "pairs":
        items = [(k, sfp(v)) for k, v in tl]
    elif mode == "nest":
        items = [(k, dict(sfp(v))) for k, v in tl]
    else:
        raise ValueError(f"Unknown mode {mode!r}")
    result = defaultdict(list)
    for tag, field in items:
        result[tag].append(field)
    return result


def record2tl(record, sfp=None, mode="field"):
    items = []
    for k, values in record.items():
        for v in values:
            items.append((k, v))

    if mode in "field":
        return items
    elif mode == "pairs":
        return [(k, sfp.unparse(*v)) for k, v in items]
    elif mode == "nest":
        return [(k, sfp.unparse(*v.items())) for k, v in items]
    else:
        raise ValueError(f"Unknown mode {mode!r}")


def nest_decode(obj, encoding):
    """Decode records in dict or tidy list format."""
    if hasattr(obj, "decode"):  # isinstance(obj, bytes)
        return obj.decode(encoding)
    if hasattr(obj, "items"):  # isinstance(obj, dict)
        return {k.decode("ascii"): nest_decode(v, encoding)
                for k, v in obj.items()}
    return [nest_decode(value, encoding) for value in obj]


def nest_encode(obj, encoding):
    """Encode records in dict or tidy list format."""
    if hasattr(obj, "encode"):  # isinstance(obj, str)
        return obj.encode(encoding)
    if hasattr(obj, "items"):  # isinstance(obj, dict)
        return {k.encode("ascii"): nest_encode(v, encoding)
                for k, v in obj.items()}
    return [nest_encode(value, encoding) for value in obj]


def utf8_fix_nest_decode(obj, encoding):
    """Decode records in dict or tidy list format
    using an hybrid strategy where a UTF-8 decoding
    is tried before the given encoding.
    """
    if hasattr(obj, "decode"):  # isinstance(obj, bytes)
        try:
            return obj.decode("utf-8")
        except UnicodeDecodeError:
            return hybrid_utf8_decode(obj, encoding)
    if hasattr(obj, "items"):  # isinstance(obj, dict)
        return {k.decode("ascii"): utf8_fix_nest_decode(v, encoding)
                for k, v in obj.items()}
    return [utf8_fix_nest_decode(value, encoding) for value in obj]


def hybrid_utf8_decode(value, encoding):
    """Decode a bytestring value that might be partially in UTF-8,
    partially in the other given encoding,
    trying to do that with UTF-8 before."""
    return "".join(
        seq.decode(enc)
        for seq, enc in zip(UTF8_MB_REGEX.split(value),
                            cycle([encoding, "utf-8"]))
        if seq
    )



#!/usr/bin/env python3
import ast
import os

import setuptools


setup_path = os.path.dirname(__file__)

# Get __version__ from the package __init__.py without importing it
with open(os.path.join(setup_path, "ioisis", "__init__.py")) as dinit:
    assignment_node = next(el for el in ast.parse(dinit.read()).body
                              if isinstance(el, ast.Assign) and
                                 el.targets[0].id == "__version__")
    version = ast.literal_eval(assignment_node.value)

with open(os.path.join(setup_path, "README.md")) as readme:
    long_description = readme.read()


setuptools.setup(
    name="ioisis",
    version=version,
    author="Danilo de Jesus da Silva Bellini",
    author_email="danilo.bellini@gmail.com",
    url="https://github.com/scieloorg/ioisis",
    description="I/O for ISIS files in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="2-clause BSD",
    packages=setuptools.find_packages(exclude=["tests"]),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "click",
        "construct",
        "JPype1",
        "ujson",
    ],
    entry_points={"console_scripts": ["ioisis = ioisis.__main__:main"]},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Java",  # Because of Bruma.jar
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Database",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Operating System :: OS Independent",
    ],
)



from functools import update_wrapper
from inspect import isgeneratorfunction, signature
import io


def should_be_file(file_argname, mode="rb"):
    """Decorator to enforce the given argument is a file-like object.
    If it isn't, it will be seen as a filename,
    and it'll be replaced by the open file with the given mode.
    """
    def decorator(func):
        if isgeneratorfunction(func):
            def wrapper(*args, **kwargs):
                bound_args = signature(func).bind(*args, **kwargs)
                file_arg = bound_args.arguments[file_argname]
                if hasattr(file_arg, "read"):  # Already a file, nothing to do
                    yield from func(*args, **kwargs)
                    return
                with open(file_arg, mode) as file_obj:  # "Cast" name to file
                    bound_args.arguments[file_argname] = file_obj
                    yield from func(*bound_args.args, **bound_args.kwargs)
        else:
            raise NotImplementedError
        return update_wrapper(wrapper, func)
    return decorator


class LineSplitError(Exception):
    pass


class LineSplittedBytesStreamWrapper:

    def __init__(self, substream, line_len, newline):
        self.substream = substream
        self.line_len = line_len
        self.newline = newline
        self.rnext_eol = line_len
        self.writing = False

    def _check_eol(self):
        if self.substream.read(len(self.newline)) != self.newline:
            raise LineSplitError("Invalid record line splitting")

    def read(self, count=None):
        result = []
        remaining = float("inf") if count is None else count
        while remaining > 0:
            expected_len = min(self.rnext_eol, remaining)
            data = self.substream.read(expected_len)
            data_len = len(data)
            result.append(data)
            remaining -= data_len
            if self.rnext_eol == data_len:
                self._check_eol()
                self.rnext_eol = self.line_len
            else:
                self.rnext_eol -= data_len
                break
        return b"".join(result)

    def write(self, data):
        self.writing = True
        result = remaining = len(data)
        while data:
            buff_len = min(self.rnext_eol, remaining)
            buff, data = (data[:buff_len], data[buff_len:])
            self.substream.write(buff)
            remaining -= buff_len
            if self.rnext_eol == buff_len:
                self.substream.write(self.newline)
                self.rnext_eol = self.line_len
            else:
                self.rnext_eol -= buff_len
                break
        return result

    def close(self):
        if self.rnext_eol != self.line_len:
            if self.writing:
                self.substream.write(self.newline)
            else:
                self._check_eol()
        self.substream = None

    tellable = lambda self: self.substream.tellable()

    def tell(self):
        line_no, col_no = divmod(self.substream.tell(),
                                 self.line_len + len(self.newline))
        return line_no * self.line_len + col_no

    seekable = lambda self: self.substream.seekable()

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            if offset < 0:
                raise ValueError("Negative offset")
            line_no, col_no = divmod(offset, self.line_len)
            line_start = line_no * (self.line_len + len(self.newline))
            self.substream.seek(line_start, whence)
            self.rnext_eol = self.line_len
            self.read(col_no)
            return offset
        elif whence == io.SEEK_CUR:
            return self.seek(self.substream.tell() + offset, io.SEEK_SET)
        elif whence == io.SEEK_END:
            if not self.finished:
                self.read()  # Just to reach the end of stream
            return self.seek(max(0, self.substream.tell() + offset),
                             io.SEEK_SET)
        raise ValueError("Invalid whence")


class TightBufferReadOnlyBytesStreamWrapper:

    def __init__(self, substream):
        self.substream = substream
        self.buffer = b""
        self.offset = 0
        self.finished = False

    def read(self, size=None):
        if size is None or size < 0:
            self.buffer += self.substream.read()
            result = self.buffer[self.offset:]
            self.finished = True
        else:
            expected_offset = self.offset + size
            missing = expected_offset - len(self.buffer)
            if missing > 0:
                self.buffer += self.substream.read(missing)
                if len(self.buffer) < expected_offset:
                    self.finished = True
            result = self.buffer[self.offset:expected_offset]
        self.offset += len(result)
        return result

    close = lambda self: None  # Required to be a file-like object
    tellable = lambda self: True
    tell = lambda self: self.offset
    seekable = lambda self: True

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            if offset < 0:
                raise ValueError("Negative offset")
            self.offset = offset
        elif whence == io.SEEK_CUR:
            self.offset = max(0, self.offset + offset)
        elif whence == io.SEEK_END:
            if not self.finished:
                self.read()  # Just to reach the end of stream
            self.offset = max(0, len(self.buffer) + offset)
        else:
            raise ValueError("Invalid whence")
        return self.offset
